/////////////////////////////////////////////////////////////////////////////
// Name:       
// Purpose:  
// Author:      Cesar Mauri Loba (cesar at crea-si dot com)
// Modified by: 
// Created:     
// Copyright:   (C) 2010 Cesar Mauri Loba - CREA Software Systems
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
/////////////////////////////////////////////////////////////////////////////
#include "oscinout.h"
#include "exceptions.h"
#include <wx/msgdlg.h>

#define ADDRESS "127.0.0.1"

COscOut::COscOut (int port)
: osc::OutboundPacketStream( m_oscBuffer, COscOut::OSC_OUTPUT_BUFFER_SIZE )
{
	assert (wxThread::IsMain());
	try
	{
		m_pTransmitSocket= new UdpTransmitSocket ( IpEndpointName( ADDRESS, port ) );
	}
	catch (std::runtime_error& e)
	{
		throw InitializationException (wxString::From8BitData(e.what()));
	}
}

COscOut::~COscOut ()
{
	Close();
}

void COscOut::Close()
{
	if (m_pTransmitSocket!= NULL) delete m_pTransmitSocket;
	m_pTransmitSocket= NULL;
}

void COscOut::Send ()
{
	assert (m_pTransmitSocket);

	if (m_pTransmitSocket== NULL)
		throw RunTimeException (_("COscOut: connection closed"));
	
	m_pTransmitSocket->Send( Data(), Size());

	Clear();
}

void COscOut::SendSimpleMessage (const char *name, float v)
{
	*this << osc::BeginBundleImmediate << osc::BeginMessage (name) 
		  << v << osc::EndMessage << osc::EndBundle;
	Send();
}

/////////////////////////////

COscIn::COscReceiverThread::COscReceiverThread (UdpListeningReceiveSocket* socket) 
: wxThread (wxTHREAD_JOINABLE) //DETACHED)
, m_socket(socket)
{
	assert (m_socket);
}

COscIn::COscReceiverThread::~COscReceiverThread()
{
	delete m_socket;
}

#define ShowError(msg,caption) \
	wxMessageDialog dlg(NULL, msg, caption); \
	dlg.ShowModal();

#define ShowErrorANSI(msg,caption) \
	wxMessageDialog dlg(NULL, wxString(msg, wxConvLocal), caption); \
	dlg.ShowModal();

wxThread::ExitCode COscIn::COscReceiverThread::Entry()
{
	wxThread::ExitCode ecode= (wxThread::ExitCode) -1;

	try	{
		m_socket->Run();
		ecode= (wxThread::ExitCode) 0;
	}
	// TODO: forward exceptions to main thread via events
	catch(InitializationException& e) {
		ShowError(e.getMessage(), _("Initialization exception"));
	}
	catch(RunTimeException& e) {
		ShowError(e.getMessage(), _("Run time exception"));
	}
	catch(BaseException& e)	{
		ShowError(e.getMessage(), _("Base exception"));
	}
	catch(exception& e)	{
		ShowErrorANSI(e.what(), _("STL exception"));
	}
	catch(...) {		
		ShowError( _("Unexpected exception"), _("Unexpected exception"));		
	}

	return ecode;
}

////////////////////////

COscIn::COscIn (int port, PacketListener *listener)
: m_thread (NULL)
, m_whoCalledSocketRun (NULL)
, m_running(false)
{
	m_socket= new UdpListeningReceiveSocket(IpEndpointName( IpEndpointName::ANY_ADDRESS, port ), listener );
}

COscIn::~COscIn()
{
	Close();
}

void COscIn::Run(bool newThread)
{
	assert (m_socket);
		
	if (m_socket== NULL)
		throw RunTimeException (_("COscIn: connection closed"));

	if (m_running)
		throw RunTimeException (_("COscIn: already running"));
	m_running= true;

	if (newThread) {
		assert (m_thread== NULL);
		// Usually we request a new thread from the main thread. Assert, just in case...
		assert (wxThread::IsMain());
       
		m_thread= new COscIn::COscReceiverThread(m_socket);
		m_thread->Create();
		m_thread->Run();
		m_whoCalledSocketRun= m_thread;
	}
	else {
		// This will block. Ensure it is not the main thread
		assert (!wxThread::IsMain());

		m_whoCalledSocketRun= wxThread::This();
		m_socket->Run();
	}
}

void COscIn::Close()
{
	if (m_socket== NULL) return;	// Already closed, ignore.
	
	// Depending on who called Run on socket Break or AsynchronousBreak should be called
	if (m_running && m_whoCalledSocketRun== wxThread::This())
	{
		// We don't allow closing from the same thread we created as we then can not wait
		// until its termination (one thread cannot wait for itself!)
		assert (m_thread== NULL);	
		m_socket->Break();	// Same thread
	}
	else 
		m_socket->AsynchronousBreak();	// Outside thread
	m_running= false;
			
	if (m_thread) { 
		//m_thread->Delete();	
		m_thread->Wait();
		delete m_thread;
		m_thread= NULL; 
	}
	else 
		delete m_socket;
	m_socket= NULL;
}