/////////////////////////////////////////////////////////////////////////////
// Name:       
// Purpose:  
// Author:      Cesar Mauri Loba (cesar at crea-si dot com)
// Modified by: 
// Created:     
// Copyright:   (C) 2010 Cesar Mauri Loba - CREA Software Systems
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
/////////////////////////////////////////////////////////////////////////////
#ifndef PUREDATAWRAPPER_H
#define PUREDATAWRAPPER_H

#include <wx/string.h>
#include <wx/event.h>
#include <wx/process.h>
#include <wx/socket.h>
#include <vector>

// Help to convert read buffer in lines
class CSimpleParser {
public:
	CSimpleParser()	{ Begin(); }
	void Begin () {	// Reset internal state
		m_inBufferSize= m_inBufferIndex= m_lineBufferSize= 0;
		m_inBufferPtr= NULL;
		m_lineBuffer[0]= 0;		
	}
	// Change buffer
	void NextBuffer (const char* buff, unsigned short size)
	{
		assert (buff);
		assert (m_inBufferSize== m_inBufferIndex);	// Ensure last buffer was fully processed
        m_inBufferPtr= buff;
		m_inBufferSize= size;
		m_inBufferIndex= 0;
	}
	// Get the next line if avaible, or NULL if no further lines available
    const char* getLine(unsigned short* lineSize= NULL)
	{
		if (m_inBufferSize== m_inBufferIndex) return NULL;	

		// Append to current line or start a new one?
		if (m_lineBuffer[m_lineBufferSize]== 0)	m_lineBufferSize= 0;	// New line

		// Copy line until new line, inBuffer end or lineBuffer full
		bool eolFound= false;
		while (m_lineBufferSize< MAXLINE-1 && m_inBufferIndex< m_inBufferSize && !eolFound)	{
			if (m_inBufferPtr[m_inBufferIndex]== '\n') {
				m_lineBuffer[m_lineBufferSize]= 0;	// EOL
				eolFound= true;				
			}
			else {
				m_lineBuffer[m_lineBufferSize]= m_inBufferPtr[m_inBufferIndex];
				m_lineBufferSize++;
			}
			m_inBufferIndex++;
		}
		if (eolFound) {
			if (lineSize) *lineSize= m_lineBufferSize;
			return m_lineBuffer;
		}
		if (m_lineBufferSize== MAXLINE-1) {
			// Not enough size to hold the entire line
			m_lineBuffer[m_lineBufferSize]= 0;
			m_lineBufferSize++;
			if (lineSize) *lineSize= m_lineBufferSize;
			return m_lineBuffer;
		}			
		return NULL;
	}
private:
	enum { MAXLINE= 512 };	
	unsigned short m_inBufferSize;
	unsigned short m_inBufferIndex;
	const char* m_inBufferPtr;
	unsigned short m_lineBufferSize;
	char m_lineBuffer[MAXLINE];	
};

class CAudioProperties
{	
public:
	int GetDelay() { return m_msDelay; }
private:
	CAudioProperties ()	{
		m_iDev1= m_iDev2= m_iDev3= m_iDev4= 0;
		m_iDev1Channels= m_iDev2Channels= m_iDev3Channels= m_iDev4Channels= 0;
		m_oDev1= m_oDev2= m_oDev3= m_oDev4= 0;
		m_oDev1Channels= m_oDev2Channels= m_oDev3Channels= m_oDev4Channels= 0;
		m_sampleRate= m_msDelay= m_canMulti= m_flongForm= 0;
	}

	std::vector<wxString> m_iDevList;
	std::vector<wxString> m_oDevList;

	// Choosen input devices
	int m_iDev1, m_iDev2, m_iDev3, m_iDev4;
	// Number of channels for each input (negative to disable input)
	int m_iDev1Channels, m_iDev2Channels, m_iDev3Channels, m_iDev4Channels;

	// Choosen output devices
	int m_oDev1, m_oDev2, m_oDev3, m_oDev4;
	// Number of channels for each output (negative to disable output)
	int m_oDev1Channels, m_oDev2Channels, m_oDev3Channels, m_oDev4Channels;

	int m_sampleRate;
	int m_msDelay;
	int m_canMulti;	// Number of max simultanious devices
	int m_flongForm;	// Undocumented

	friend class CPureDataWrapper;

};

class CPureDataWrapper : public wxEvtHandler
{
public:
	// Singleton
	static CPureDataWrapper* getInstance();
	static void destroyInstance();
	
	void StartPD ();
	void StopPD ();
	const wxString& getVersion() { return m_pdVersion; }
	// Opens a patch and returns the pathId or raises an exception if failed
	wxString OpenPatch (const wxString& file);
	void ClosePatch (const wxString& patchId);
	void StartDSP ();
	void StopDSP ();

	long getCurrentAPI() const { return m_whichAPI; }
	void setCurrentAPI(long apiNum);

	CAudioProperties* getAudioProperties ();
	void setAudioProperties (const CAudioProperties& ap, bool savePreferences= false);

	void SetIntelligentASIOConfig (unsigned long delay, bool savePreferences= false);
	
	bool getDebugGUIMode () const { return m_debugGUIMode; }
	void setDebugGUIMode (bool enable);

	void EnableLog (FILE *stream= NULL);
	void DisableLog ();
	bool IsLogEnabled () { return m_logStream!= NULL; }
private:
	// Construction
	static CPureDataWrapper* g_single;
	CPureDataWrapper();
	~CPureDataWrapper(void);	
	
	// Enums
	enum EPdWrapperStatus {NOT_RUNNING= 0, WAIT_INCOMING_CONNECTION, INIT1, INIT2, INIT3, RUNNING, WAIT_TERMINATE};
	enum EParserStatus { IGNORE_INPUT= 0, WAIT_ACK, WAIT_OPEN_PATCH, WAIT_CLOSE_PATCH, AUDIO_PROP_IN1, AUDIO_PROP_IN2,
						 AUDIO_PROP_OUT1, AUDIO_PROP_OUT2, AUDIO_PROP_VALUES, WAIT_WHICH_API };
	
	// Auxiliary
    void ParseInput (const char* buffer, wxUint32 size);
	wxString CorrectFilePath (const wxString& fp) const;
	// True is state changed to e, false if timeout
	bool WaitWhileParserStatusIs (EParserStatus e, int time_cs= 50);
	// True is state, false if timeout
	bool WaitWhileParserStatusIsNot (EParserStatus e, int time_cs= 50);
	void SendMessageToPD (const wxString& msg);
	void ManageAudioOptionsDialog (const wxString& msg);

	void LaunchPD (const wxString& params);
	
	// Check whether the call is not reentrant and PD is running, used by
	// almost all public methods. Returns false if is not possible to run
	// the method now (i.e. the caller should return)
	bool CommonChecks ();
	
	typedef struct { wxString apiName; long apiNum; } TApiDesc;
	std::vector<TApiDesc> m_apiList;

	CAudioProperties m_audioProperties;

	// Attributes
	long m_msAudioBuffer;
	bool m_debugGUIMode;
	wxString m_patchName;
	wxProcess m_process;
	long m_pid;
	wxSocketBase* m_pdConnection;
	EPdWrapperStatus m_status;
	EParserStatus m_parserStatus;
	CSimpleParser m_parser;
	wxString m_pdVersion;
	long m_whichAPI;
	FILE* m_logStream;
	bool m_ownConsole;
	
	bool m_entry;
	wxString m_tmpExchange;
	bool m_tmpError;

	// Events
	void OnProcessTerm(wxProcessEvent& event);
	void OnSocketEvent(wxSocketEvent& event);


	/*
	void RunPatch(void);
	long FindASIOInputDevice(const wxArrayString& pdOutput);
	long FindASIOOutputDevice(const wxArrayString& pdOutput);
	long FindASIODevice (const wxArrayString& pdOutput, const wxString& header);
	*/


	// any class wishing to process wxWidgets events must use this macro
	DECLARE_EVENT_TABLE()
};

// Auxiliary class to manage automatic PD start/stop when opening/closing patchs
class CPureDataSimpleController
{
public:
	static CPureDataSimpleController* getInstance();
	static void destroyInstance();

	wxString OpenPatch (const wxString& file);
	void ClosePatch (const wxString& patchId);
private:
	CPureDataSimpleController();
	wxArrayString m_openPatchs;
	static CPureDataSimpleController* g_single;
};

#endif