/////////////////////////////////////////////////////////////////////////////
// Name:       
// Purpose:  
// Author:      Cesar Mauri Loba (cesar at crea-si dot com)
// Modified by: 
// Created:     
// Copyright:   (C) 2010 Cesar Mauri Loba - CREA Software Systems
// 
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
/////////////////////////////////////////////////////////////////////////////
#pragma once

#include <assert.h>
#include <math.h>
#include "exceptions.h"

template <typename T> class CValueRange
{
public:
	CValueRange(T value, T min= 0, T max= 0, T defaul= 0) //, T step= 1)
	{
		m_min= min;
		m_max= max;
		if (defaul< min || defaul> max)
			throw ValueOutOfRangeException(_T("CValueRange"));	
		m_default= defaul;
		setValue (value);
	}

	void setValue (T val)
	{
		if (val< m_min || val> m_max) 
			throw ValueOutOfRangeException(_T("CValueRange"));
		m_value= val;
	}

	T getDefault()
	{
		return m_default;
	}

	T getValue ()
	{
		return m_value;
	}

	T getMin()
	{
		return m_min;
	}

	T getMax()
	{
		return m_max;
	}

private:
	T m_value;
	T m_min;
	T m_max;
	T m_default;
};

typedef CValueRange<int> CValueRangeI;
typedef CValueRange<unsigned short> CValueRangeUS;

//
// This class is intended for helping mapping between linear to exp scales.
// The main purpose is to allow linear sliders produce a exp range of values
// The following function is used:
//
//  y= A*e^(x*B)-A+C  (defined for x>= 0)
//
// Where 
//  *	x is the input value (linear)
//	*	y is the output value (exp)
//	*	A is the "speed" constant (for large values the above function grows faster)
//		For convenience A=e^a where "a" is a parameter set by the user and can be 
//		either positive or negative.
//	*	B and C are constants determined by two points the function should cross
//			p1=(min_linear>= 0, min_exp)  
//				and  
//			p2=(max_linear> 0, maxexp> min_exp)
//
// For simplicity min_linear= 0. This yields the following solutions for
// the equation system:
//
//		C= min_exp	
//		B= (1/max_lin) * ln ((A+max_exp-min_exp)/A)
//
// The inverse function is:
//
//  x= (1/B) * ln ((y+A-C)/A)
//

class CValueRangeFExp
{
public:
	CValueRangeFExp (float min_linear, float min_exp, float max_linear, float max_exp, float default_exp, float grow= 0)
	{
		SetParams (min_linear, min_exp, max_linear, max_exp, default_exp, grow);
		m_expValue= default_exp;
	}

	float GetMinLinear () { return 0.0f; }
	float GetMaxLinear () { return ToLinear(m_maxExp); }
	float GetDefaultLinear () { return ToLinear(m_defaultExp); }
	
	float GetMinExp () { return m_C; }
	float GetMaxExp () { return m_maxExp; }
	float GetDefaultExp () { return m_defaultExp; }

	float GetLinearValue () { return ToLinear(m_expValue); }
	void SetLinearValue (float lin) 
	{
		assert (lin>= GetMinLinear ());
		assert (lin<= GetMaxLinear ());

		// Only assert these conditions but do not 
		// throw exceptions as it is not dangerous
		m_expValue= ToExp(lin);		
	}

	float GetExpValue () { return m_expValue; }
	void SetExpValue (float expv) 
	{
		assert (expv>= GetMinExp());
		assert (expv<= GetMaxExp ());

		// Raise an exception only if the new value would
		// produce further analytical errors
		if (!(expv>= GetMinExp())) 
			throw ValueOutOfRangeException(_T("CValueRangeFexp: !(expv>= GetMinExp())"));
		m_expValue= expv;
	}
private:
	void SetParams (float min_linear, float min_exp, float max_linear, float max_exp, float default_exp, float grow)
	{
		assert (min_linear== 0);
		if (min_linear!= 0) throw ValueOutOfRangeException(_T("CValueRangeFexp: wrong min_linear"));
        assert (min_exp>= 0);
		if (!(min_exp>= 0)) throw ValueOutOfRangeException(_T("CValueRangeFexp: wrong min_exp"));
		assert (max_linear> min_linear);
		if (!(max_linear> min_linear)) throw ValueOutOfRangeException(_T("CValueRangeFexp: wrong max_linear"));
		assert (max_exp> min_exp);
		if (!(max_exp> min_exp)) throw ValueOutOfRangeException(_T("CValueRangeFexp: wrong max_exp"));
		assert (default_exp>= min_exp && default_exp<= max_exp);
		if (!(default_exp>= min_exp && default_exp<= max_exp))
			throw ValueOutOfRangeException(_T("CValueRangeFexp: wrong default_exp"));
			
		m_A= exp (grow);
		m_C= min_exp;
		m_B= (1.0f / max_linear) * logf ((m_A + max_exp - m_C) / m_A);

		m_maxExp= max_exp;
		m_defaultExp= default_exp;		
	}

	float ToExp (float x)
	{
		return m_A * exp (x*m_B) - m_A + m_C;
	}

	float ToLinear (float y)
	{
		return (1.0f/m_B) * logf ((y+m_A-m_C)/m_A);
	}

private:
	float m_A;
	float m_B;
	float m_C;	// Equal to min_exp

	float m_maxExp;
	float m_defaultExp;

	float m_expValue;
};